--- Schematic serialization and deserialiation.
-- @module worldedit.serialization

worldedit.LATEST_SERIALIZATION_VERSION = 5
local LATEST_SERIALIZATION_HEADER = worldedit.LATEST_SERIALIZATION_VERSION .. ":"


--[[
Serialization version history:
  1: Original format.  Serialized Lua table with a weird linked format...
  2: Position and node seperated into sub-tables in fields `1` and `2`.
  3: List of nodes, one per line, with fields seperated by spaces.
      Format: <X> <Y> <Z> <Name> <Param1> <Param2>
  4: Serialized Lua table containing a list of nodes with `x`, `y`, `z`,
      `name`, `param1`, `param2`, and `meta` fields.
  5: Added header and made `param1`, `param2`, and `meta` fields optional.
      Header format: <Version>,<ExtraHeaderField1>,...:<Content>
--]]


--- Reads the header of serialized data.
-- @param value Serialized WorldEdit data.
-- @return The version as a positive natural number, or 0 for unknown versions.
-- @return Extra header fields as a list of strings, or nil if not supported.
-- @return Content (data after header).
function worldedit.read_header(value)
    if value:find("^[0-9]+[%-:]") then
        local header_end = value:find(":", 1, true)
        local header = value:sub(1, header_end - 1):split(",")
        local version = tonumber(header[1])
        table.remove(header, 1)
        local content = value:sub(header_end + 1)
        return version, header, content
    end
    -- Old versions that didn't include a header with a version number
    if value:find("([+-]?%d+)%s+([+-]?%d+)%s+([+-]?%d+)") and not value:find("%{") then -- List format
        return 3, nil, value
    elseif value:find("^[^\"']+%{%d+%}") then
        if value:find("%[\"meta\"%]") then -- Meta flat table format
            return 2, nil, value
        end
        return 1, nil, value -- Flat table format
    elseif value:find("%{") then -- Raw nested table format
        return 4, nil, value
    end
    return nil
end

-- ===================================================================

local function Set (list)
    local set = {}
    for _, l in ipairs (list) do set [l] = true end
    return set
end
-- local myset = Set{"while", "end", "function", "local", }

-- ===================================================================

--- Converts the region defined by positions `pos1` and `pos2`
-- into a single string.
-- @return The serialized data.
-- @return The number of nodes serialized.

-- ===================================================================

local small_bowl = 0

-- ===================================================================

-- Nodes which use "meta" should be listed in "needwe", but this isn't
-- essential.  "needwe" isn't presently used,  but it might be used in
-- the future.

local needwe     = Set{
    "default:bookshelf"            ,
    "default:chest"                ,
    "default:chest_locked"         ,
    "default:furnace"              ,
    "default:sign_wall_steel"      ,
    "default:sign_wall_wood"       ,
    "locked_sign:sign_wall_locked" ,
    "signs:sign_hanging"           ,
    "signs:sign_wall_blue"         ,
    "signs:sign_wall_brown"        ,
    "signs:sign_wall_green"        ,
    "signs:sign_wall_orange"       ,
    "signs:sign_wall_red"          ,
    "signs:sign_wall_white_black"  ,
    "signs:sign_wall_white_red"    ,
    "signs:sign_wall_yellow"       ,
    "signs:sign_yard"              ,
    "vessels:shelf"                ,
}

-- ===================================================================

-- "rejectwe" should list the majority  of common nodes  that do *not*
-- contain meta.

-- The nodes listed in "rejectwe" may use "facedir",  but are believed
-- not to use "meta" per se. Nodes which turn out to use "meta" should
-- be removed from the list.

local rejectwe   = Set{
    "air"                                                    ,
    "ignore"                                                 ,

    "abriglass:clear_glass"                                  ,
    "abriglass:ghost_crystal"                                ,
    "abriglass:glass_light_blue"                             ,
    "abriglass:glass_light_green"                            ,
    "abriglass:glass_light_red"                              ,
    "abriglass:glass_light_yellow"                           ,
    "abriglass:hidden_light"                                 ,
    "abriglass:oneway_glass_dark"                            ,
    "abriglass:oneway_glass_pale"                            ,
    "abriglass:oneway_wall_dark"                             ,
    "abriglass:oneway_wall_pale"                             ,
    "abriglass:stained_glass_frosted"                        ,

    "asteroid:atmos"                                         ,
    "asteroid:cobble"                                        ,
    "asteroid:copperore"                                     ,
    "asteroid:dust"                                          ,
    "asteroid:gravel"                                        ,
    "asteroid:ironore"                                       ,
    "asteroid:meteorite_block"                               ,

    "bakedclay:black"                                        ,
    "bakedclay:delphinium"                                   ,
    "bakedclay:lazarus"                                      ,
    "bakedclay:mannagrass"                                   ,
    "bakedclay:red"                                          ,
    "bakedclay:thistle"                                      ,
    "bakedclay:white"                                        ,

    "beds:bed_bottom"                                        ,
    "beds:bed_top"                                           ,
    "beds:fancy_bed_bottom"                                  ,
    "beds:fancy_bed_top"                                     ,

    "birthstones:alexandriteblock"                           ,
    "birthstones:alexandriteore"                             ,
    "birthstones:amethystblock"                              ,
    "birthstones:amethystore"                                ,
    "birthstones:aquamarineblock"                            ,
    "birthstones:aquamarineore"                              ,
    "birthstones:diamondblock"                               ,
    "birthstones:diamondore"                                 ,
    "birthstones:emeraldblock"                               ,
    "birthstones:emeraldore"                                 ,
    "birthstones:garnetblock"                                ,
    "birthstones:garnetore"                                  ,
    "birthstones:opalblock"                                  ,
    "birthstones:opalore"                                    ,
    "birthstones:peridotblock"                               ,
    "birthstones:peridotore"                                 ,
    "birthstones:rubyblock"                                  ,
    "birthstones:rubyore"                                    ,
    "birthstones:sapphireblock"                              ,
    "birthstones:sapphireore"                                ,
    "birthstones:topazblock"                                 ,
    "birthstones:topazore"                                   ,
    "birthstones:zirconblock"                                ,
    "birthstones:zirconore"                                  ,

    "bluelava:lava_flowing"                                  ,
    "bluelava:lava_source"                                   ,

    "building_blocks:Adobe"                                  ,
    "building_blocks:brobble_spread"                         ,
    "building_blocks:BWtile"                                 ,
    "building_blocks:fakegrass"                              ,
    "building_blocks:grate"                                  ,
    "building_blocks:gravel_spread"                          ,
    "building_blocks:hardwood"                               ,

    "building_blocks:Marble"                                 ,
    "building_blocks:Roofing"                                ,
    "building_blocks:slab_marble"                            ,
    "building_blocks:smoothglass"                            ,
    "building_blocks:Tar"                                    ,
    "building_blocks:Tarmac_spread"                          ,
    "building_blocks:terrycloth_towel"                       ,
    "building_blocks:woodglass"                              ,

    "caverealms:coal_dust"                                   ,
    "caverealms:constant_flame"                              ,
    "caverealms:dm_statue"                                   ,
    "caverealms:fire_vine"                                   ,
    "caverealms:fungus"                                      ,
    "caverealms:glow_amethyst"                               ,
    "caverealms:glow_amethyst_ore"                           ,
    "caverealms:glow_crystal"                                ,
    "caverealms:glow_emerald"                                ,
    "caverealms:glow_emerald_ore"                            ,
    "caverealms:glow_gem"                                    ,
    "caverealms:glow_gem_2"                                  ,
    "caverealms:glow_gem_3"                                  ,
    "caverealms:glow_gem_4"                                  ,
    "caverealms:glow_gem_5"                                  ,
    "caverealms:glow_mese"                                   ,
    "caverealms:glow_obsidian"                               ,
    "caverealms:glow_obsidian_2"                             ,
    "caverealms:glow_obsidian_glass"                         ,
    "caverealms:glow_ore"                                    ,
    "caverealms:glow_ruby"                                   ,
    "caverealms:glow_ruby_ore"                               ,
    "caverealms:glow_worm"                                   ,
    "caverealms:glow_worm_green"                             ,
    "caverealms:hanging_thin_ice"                            ,
    "caverealms:hot_cobble"                                  ,
    "caverealms:icicle_down"                                 ,
    "caverealms:icicle_up"                                   ,
    "caverealms:mushroom_cap"                                ,
    "caverealms:mushroom_gills"                              ,
    "caverealms:mushroom_sapling"                            ,
    "caverealms:mushroom_stem"                               ,
    "caverealms:mycena"                                      ,
    "caverealms:s_chest"                                     ,
    "caverealms:s_fortress"                                  ,
    "caverealms:s_fountain"                                  ,
    "caverealms:salt_crystal"                                ,
    "caverealms:salt_gem"                                    ,
    "caverealms:spike"                                       ,
    "caverealms:stone_with_algae"                            ,
    "caverealms:stone_with_lichen"                           ,
    "caverealms:stone_with_moss"                             ,
    "caverealms:stone_with_salt"                             ,
    "caverealms:thin_ice"                                    ,

    "chains:chain"                                           ,
    "chains:chain_brass"                                     ,
    "chains:chain_top"                                       ,
    "chains:chain_top_brass"                                 ,
    "chains:chandelier"                                      ,
    "chains:chandelier_brass"                                ,

    "codersea:dirt"                                          ,
    "codersea:dirtalgae"                                     ,
    "codersea:dirtgrow"                                      ,
    "codersea:dirtkelp"                                      ,
    "codersea:dirtseagrass"                                  ,

    "codersea:ethereal_coral2"                               ,
    "codersea:ethereal_coral3"                               ,
    "codersea:ethereal_coral4"                               ,
    "codersea:ethereal_coral5"                               ,
    "codersea:ethereal_seaweed"                              ,

    "codersea:kelpbrown"                                     ,
    "codersea:kelpbrownmiddle"                               ,
    "codersea:kelpgreen"                                     ,
    "codersea:kelpgreenmiddle"                               ,

    "codersea:sand"                                          ,
    "codersea:sandalgae"                                     ,
    "codersea:sandgrow"                                      ,
    "codersea:sandkelp"                                      ,
    "codersea:sandseagrass"                                  ,

    "codersea:seabrick"                                      ,
    "codersea:seabrick_aqua"                                 ,
    "codersea:seabrick_cyan"                                 ,
    "codersea:seabrick_lime"                                 ,
    "codersea:seabrick_magenta"                              ,
    "codersea:seabrick_redviolet"                            ,
    "codersea:seabrick_skyblue"                              ,

    "codersea:seacobble"                                     ,
    "codersea:seacobble_aqua"                                ,
    "codersea:seacobble_cyan"                                ,
    "codersea:seacobble_lime"                                ,
    "codersea:seacobble_magenta"                             ,
    "codersea:seacobble_redviolet"                           ,
    "codersea:seacobble_skyblue"                             ,

    "codersea:seaglass"                                      ,
    "codersea:seaglass_black"                                ,
    "codersea:seaglass_black_off"                            ,
    "codersea:seaglass_blue"                                 ,
    "codersea:seaglass_blue_off"                             ,
    "codersea:seaglass_off"                                  ,
    "codersea:seaglass_red"                                  ,
    "codersea:seaglass_red_off"                              ,
    "codersea:seaglass_white"                                ,
    "codersea:seaglass_white_off"                            ,
    "codersea:seaglass_yellow"                               ,
    "codersea:seaglass_yellow_off"                           ,

    "codersea:seagrassgreen"                                 ,
    "codersea:seagrassred"                                   ,

    "codersea:seagravel"                                     ,
    "codersea:seagravel_aqua"                                ,
    "codersea:seagravel_cyan"                                ,
    "codersea:seagravel_lime"                                ,
    "codersea:seagravel_magenta"                             ,
    "codersea:seagravel_redviolet"                           ,
    "codersea:seagravel_skyblue"                             ,

    "codersea:seaplantsdirtkelpbrown"                        ,
    "codersea:seaplantsdirtkelpgreen"                        ,
    "codersea:seaplantsdirtseagrassgreen"                    ,
    "codersea:seaplantsdirtseagrassred"                      ,
    "codersea:seaplantssandkelpbrown"                        ,
    "codersea:seaplantssandkelpgreen"                        ,
    "codersea:seaplantssandseagrassgreen"                    ,
    "codersea:seaplantssandseagrassred"                      ,

    "codersea:seastone"                                      ,
    "codersea:seastone_aqua"                                 ,
    "codersea:seastone_cyan"                                 ,
    "codersea:seastone_lime"                                 ,
    "codersea:seastone_magenta"                              ,
    "codersea:seastone_redviolet"                            ,
    "codersea:seastone_skyblue"                              ,

    "codersea:shine_slab_seaglass"                           ,
    "codersea:shine_slab_seaglass_black"                     ,
    "codersea:shine_slab_seaglass_black_off"                 ,
    "codersea:shine_slab_seaglass_blue"                      ,
    "codersea:shine_slab_seaglass_blue_off"                  ,
    "codersea:shine_slab_seaglass_off"                       ,
    "codersea:shine_slab_seaglass_red"                       ,
    "codersea:shine_slab_seaglass_red_off"                   ,
    "codersea:shine_slab_seaglass_white"                     ,
    "codersea:shine_slab_seaglass_white_off"                 ,
    "codersea:shine_slab_seaglass_yellow"                    ,
    "codersea:shine_slab_seaglass_yellow_off"                ,

    "codersea:shine_stair_seaglass"                          ,
    "codersea:shine_stair_seaglass_black"                    ,
    "codersea:shine_stair_seaglass_black_off"                ,
    "codersea:shine_stair_seaglass_blue"                     ,
    "codersea:shine_stair_seaglass_blue_off"                 ,
    "codersea:shine_stair_seaglass_off"                      ,
    "codersea:shine_stair_seaglass_red"                      ,
    "codersea:shine_stair_seaglass_red_off"                  ,
    "codersea:shine_stair_seaglass_white"                    ,
    "codersea:shine_stair_seaglass_white_off"                ,
    "codersea:shine_stair_seaglass_yellow"                   ,
    "codersea:shine_stair_seaglass_yellow_off"               ,

    "codersea:water_flowing"                                 ,
    "codersea:water_source"                                  ,

    "computer:piepad"                                        ,
    "computer:printer"                                       ,
    "computer:router"                                        ,
    "computer:server"                                        ,
    "computer:server_on"                                     ,
    "computer:slaystation"                                   ,
    "computer:tetris_arcade"                                 ,
    "computer:tower"                                         ,
    "computer:vanio"                                         ,
    "computer:vanio_off"                                     ,

    "default:acacia_bush_leaves"                             ,
    "default:acacia_bush_sapling"                            ,
    "default:acacia_bush_stem"                               ,
    "default:acacia_leaves"                                  ,
    "default:acacia_sapling"                                 ,
    "default:acacia_tree"                                    ,
    "default:acacia_wood"                                    ,

    "default:acorn"                                          ,
    "default:apple"                                          ,

    "default:aspen_leaves"                                   ,
    "default:aspen_sapling"                                  ,
    "default:aspen_tree"                                     ,
    "default:aspen_wood"                                     ,
    "default:brick"                                          ,
    "default:bronzeblock"                                    ,

    "default:bush_leaves"                                    ,
    "default:bush_sapling"                                   ,
    "default:bush_stem"                                      ,
    "default:cactus"                                         ,
    "default:clay"                                           ,
    "default:cloud"                                          ,
    "default:coalblock"                                      ,
    "default:cobble"                                         ,
    "default:cobble_microslab"                               ,
    "default:copperblock"                                    ,
    "default:coral_brown"                                    ,
    "default:coral_orange"                                   ,
    "default:coral_skeleton"                                 ,
    "default:desert_cobble"                                  ,

    "default:desert_sand"                                    ,
    "default:desert_sandstone"                               ,
    "default:desert_sandstone_block"                         ,
    "default:desert_sandstone_brick"                         ,
    "default:desert_stone"                                   ,
    "default:desert_stone_block"                             ,
    "default:desert_stonebrick"                              ,
    "default:diamondblock"                                   ,

    "default:dirt"                                           ,
    "default:dirt_with_dry_grass"                            ,
    "default:dirt_with_grass"                                ,
    "default:dirt_with_grass_footsteps"                      ,
    "default:dirt_with_rainforest_litter"                    ,
    "default:dirt_with_snow"                                 ,

    "default:dry_grass_1"                                    ,
    "default:dry_grass_2"                                    ,
    "default:dry_grass_3"                                    ,
    "default:dry_grass_4"                                    ,
    "default:dry_grass_5"                                    ,
    "default:dry_shrub"                                      ,

    "default:fence_junglewood"                               ,
    "default:fence_wood"                                     ,
    "default:glass"                                          ,
    "default:goldblock"                                      ,
    "default:goldblock_microslab"                            ,

    "default:grass_1"                                        ,
    "default:grass_2"                                        ,
    "default:grass_3"                                        ,
    "default:grass_4"                                        ,
    "default:grass_5"                                        ,

    "default:gravel"                                         ,
    "default:ice"                                            ,

    "default:junglegrass"                                    ,
    "default:jungleleaves"                                   ,
    "default:junglesapling"                                  ,
    "default:jungletree"                                     ,
    "default:junglewood"                                     ,

    "default:ladder_steel"                                   ,
    "default:ladder_wood"                                    ,

    "default:lava"                                           ,
    "default:lava_flowing"                                   ,
    "default:lava_source"                                    ,
    "default:leaves"                                         ,
    "default:mese"                                           ,
    "default:meselamp"                                       ,
    "default:mese_post_light"                                ,
    "default:mossycobble"                                    ,

    "default:oak_leaves"                                     ,
    "default:oak_sapling"                                    ,
    "default:oak_trunk"                                      ,
    "default:oak_wood"                                       ,

    "default:obsidian"                                       ,
    "default:obsidian_block"                                 ,
    "default:obsidianbrick"                                  ,
    "default:obsidian_glass"                                 ,
    "default:papyrus"                                        ,

    "default:pine_needles"                                   ,
    "default:pine_sapling"                                   ,
    "default:pine_tree"                                      ,
    "default:pine_wood"                                      ,

    "default:river_water_flowing"                            ,
    "default:river_water_source"                             ,

    "default:rubber_leaves"                                  ,
    "default:rubber_sapling"                                 ,
    "default:rubber_trunk"                                   ,
    "default:rubber_wood"                                    ,
    "default:sand"                                           ,

    "default:sandstone"                                      ,
    "default:sandstone_block"                                ,
    "default:sandstonebrick"                                 ,

    "default:sapling"                                        ,

    "default:sequoia_leaves"                                 ,
    "default:sequoia_sapling"                                ,
    "default:sequoia_trunk"                                  ,
    "default:sequoia_wood"                                   ,

    "default:silver_sand"                                    ,
    "default:silver_sandstone"                               ,
    "default:silver_sandstone_block"                         ,
    "default:silver_sandstone_brick"                         ,
    "default:snow"                                           ,
    "default:snowblock"                                      ,

    "default:spruce_cone"                                    ,
    "default:spruce_leaves"                                  ,
    "default:spruce_sapling"                                 ,
    "default:spruce_spruce"                                  ,
    "default:spruce_trunk"                                   ,
    "default:steelblock"                                     ,

    "default:stone"                                          ,
    "default:stone_block"                                    ,
    "default:stonebrick"                                     ,
    "default:stonebrick_microslab"                           ,
    "default:stone_innerstair"                               ,
    "default:stone_microslab"                                ,
    "default:stone_panel"                                    ,
    "default:stone_with_coal"                                ,
    "default:stone_with_copper"                              ,
    "default:stone_with_diamond"                             ,
    "default:stone_with_gold"                                ,
    "default:stone_with_iron"                                ,
    "default:stone_with_mese"                                ,
    "default:stone_with_tin"                                 ,
    "default:tinblock"                                       ,
    "default:torch"                                          ,
    "default:torch_ceiling"                                  ,
    "default:torch_wall"                                     ,
    "default:tree"                                           ,

    "default:water"                                          ,
    "default:water_flowing"                                  ,
    "default:water_source"                                   ,

    "default:wood"                                           ,
    "default:wood_leaves"                                    ,
    "default:wood_sapling"                                   ,
    "default:wood_trunk"                                     ,
    "default:wood_wood"                                      ,

    "ethereal:bamboo"                                        ,
    "ethereal:bamboo_block"                                  ,
    "ethereal:bamboo_dirt"                                   ,
    "ethereal:bamboo_floor"                                  ,
    "ethereal:bamboo_leaves"                                 ,
    "ethereal:bamboo_sprout"                                 ,

    "ethereal:banana"                                        ,
    "ethereal:banana_tree_sapling"                           ,
    "ethereal:banana_trunk"                                  ,
    "ethereal:banana_wood"                                   ,

    "ethereal:bananaleaves"                                  ,
    "ethereal:big_tree_sapling"                              ,

    "ethereal:birch_leaves"                                  ,
    "ethereal:birch_sapling"                                 ,
    "ethereal:birch_trunk"                                   ,
    "ethereal:birch_wood"                                    ,

    "ethereal:bush"                                          ,
    "ethereal:bush2"                                         ,
    "ethereal:bush3"                                         ,
    "ethereal:candle"                                        ,
    "ethereal:coconut"                                       ,
    "ethereal:cold_dirt"                                     ,

    "ethereal:coral2"                                        ,
    "ethereal:coral3"                                        ,
    "ethereal:coral4"                                        ,
    "ethereal:coral5"                                        ,

    "ethereal:crystal_block"                                 ,
    "ethereal:crystal_dirt"                                  ,
    "ethereal:crystal_moss"                                  ,
    "ethereal:crystal_spike"                                 ,
    "ethereal:crystalgrass"                                  ,
    "ethereal:door_sakura_a"                                 ,
    "ethereal:door_sakura_b"                                 ,
    "ethereal:dry_dirt"                                      ,
    "ethereal:dry_shrub"                                     ,

    "ethereal:fence_banana"                                  ,
    "ethereal:fence_birch"                                   ,
    "ethereal:fence_frostwood"                               ,
    "ethereal:fence_mushroom"                                ,
    "ethereal:fence_olive"                                   ,
    "ethereal:fence_palm"                                    ,
    "ethereal:fence_redwood"                                 ,
    "ethereal:fence_sakura"                                  ,
    "ethereal:fence_scorched"                                ,
    "ethereal:fence_willow"                                  ,
    "ethereal:fence_yellowwood"                              ,

    "ethereal:fern"                                          ,
    "ethereal:fiery_dirt"                                    ,
    "ethereal:fiery_moss"                                    ,
    "ethereal:fire_flower"                                   ,
    "ethereal:firethorn"                                     ,
    "ethereal:frost_leaves"                                  ,
    "ethereal:frost_tree"                                    ,
    "ethereal:frost_tree_sapling"                            ,
    "ethereal:frost_wood"                                    ,

    "ethereal:glostone"                                      ,
    "ethereal:golden_apple"                                  ,
    "ethereal:gray_dirt"                                     ,
    "ethereal:gray_moss"                                     ,
    "ethereal:green_moss"                                    ,
    "ethereal:grove_dirt"                                    ,

    "ethereal:icebrick"                                      ,
    "ethereal:illumishroom"                                  ,
    "ethereal:illumishroom2"                                 ,
    "ethereal:illumishroom3"                                 ,
    "ethereal:jungle_dirt"                                   ,
    "ethereal:lemon"                                         ,
    "ethereal:lemon_leaves"                                  ,
    "ethereal:lemon_tree_sapling"                            ,
    "ethereal:lightstring"                                   ,

    "ethereal:micro_bamboo_wood"                             ,
    "ethereal:micro_bamboo_wood_1"                           ,
    "ethereal:micro_bamboo_wood_12"                          ,
    "ethereal:micro_bamboo_wood_14"                          ,
    "ethereal:micro_bamboo_wood_15"                          ,
    "ethereal:micro_bamboo_wood_2"                           ,
    "ethereal:micro_bamboo_wood_4"                           ,
    "ethereal:micro_banana_wood"                             ,
    "ethereal:micro_banana_wood_1"                           ,
    "ethereal:micro_banana_wood_12"                          ,
    "ethereal:micro_banana_wood_14"                          ,
    "ethereal:micro_banana_wood_15"                          ,
    "ethereal:micro_banana_wood_2"                           ,
    "ethereal:micro_banana_wood_4"                           ,
    "ethereal:micro_birch_wood"                              ,
    "ethereal:micro_birch_wood_1"                            ,
    "ethereal:micro_birch_wood_12"                           ,
    "ethereal:micro_birch_wood_14"                           ,
    "ethereal:micro_birch_wood_15"                           ,
    "ethereal:micro_birch_wood_2"                            ,
    "ethereal:micro_birch_wood_4"                            ,
    "ethereal:micro_crystal_block"                           ,
    "ethereal:micro_crystal_block_1"                         ,
    "ethereal:micro_crystal_block_12"                        ,
    "ethereal:micro_crystal_block_14"                        ,
    "ethereal:micro_crystal_block_15"                        ,
    "ethereal:micro_crystal_block_2"                         ,
    "ethereal:micro_crystal_block_4"                         ,
    "ethereal:micro_dry_dirt"                                ,
    "ethereal:micro_dry_dirt_1"                              ,
    "ethereal:micro_dry_dirt_12"                             ,
    "ethereal:micro_dry_dirt_14"                             ,
    "ethereal:micro_dry_dirt_15"                             ,
    "ethereal:micro_dry_dirt_2"                              ,
    "ethereal:micro_dry_dirt_4"                              ,
    "ethereal:micro_frost_wood"                              ,
    "ethereal:micro_frost_wood_1"                            ,
    "ethereal:micro_frost_wood_12"                           ,
    "ethereal:micro_frost_wood_14"                           ,
    "ethereal:micro_frost_wood_15"                           ,
    "ethereal:micro_frost_wood_2"                            ,
    "ethereal:micro_frost_wood_4"                            ,
    "ethereal:micro_icebrick"                                ,
    "ethereal:micro_icebrick_1"                              ,
    "ethereal:micro_icebrick_12"                             ,
    "ethereal:micro_icebrick_14"                             ,
    "ethereal:micro_icebrick_15"                             ,
    "ethereal:micro_icebrick_2"                              ,
    "ethereal:micro_icebrick_4"                              ,
    "ethereal:micro_mushroom"                                ,
    "ethereal:micro_mushroom_1"                              ,
    "ethereal:micro_mushroom_12"                             ,
    "ethereal:micro_mushroom_14"                             ,
    "ethereal:micro_mushroom_15"                             ,
    "ethereal:micro_mushroom_2"                              ,
    "ethereal:micro_mushroom_4"                              ,
    "ethereal:micro_mushroom_trunk"                          ,
    "ethereal:micro_mushroom_trunk_1"                        ,
    "ethereal:micro_mushroom_trunk_12"                       ,
    "ethereal:micro_mushroom_trunk_14"                       ,
    "ethereal:micro_mushroom_trunk_15"                       ,
    "ethereal:micro_mushroom_trunk_2"                        ,
    "ethereal:micro_mushroom_trunk_4"                        ,
    "ethereal:micro_olive_wood"                              ,
    "ethereal:micro_olive_wood_1"                            ,
    "ethereal:micro_olive_wood_12"                           ,
    "ethereal:micro_olive_wood_14"                           ,
    "ethereal:micro_olive_wood_15"                           ,
    "ethereal:micro_olive_wood_2"                            ,
    "ethereal:micro_olive_wood_4"                            ,
    "ethereal:micro_palm_wood"                               ,
    "ethereal:micro_palm_wood_1"                             ,
    "ethereal:micro_palm_wood_12"                            ,
    "ethereal:micro_palm_wood_14"                            ,
    "ethereal:micro_palm_wood_15"                            ,
    "ethereal:micro_palm_wood_2"                             ,
    "ethereal:micro_palm_wood_4"                             ,
    "ethereal:micro_redwood_wood"                            ,
    "ethereal:micro_redwood_wood_1"                          ,
    "ethereal:micro_redwood_wood_12"                         ,
    "ethereal:micro_redwood_wood_14"                         ,
    "ethereal:micro_redwood_wood_15"                         ,
    "ethereal:micro_redwood_wood_2"                          ,
    "ethereal:micro_redwood_wood_4"                          ,
    "ethereal:micro_sakura_wood"                             ,
    "ethereal:micro_sakura_wood_1"                           ,
    "ethereal:micro_sakura_wood_12"                          ,
    "ethereal:micro_sakura_wood_14"                          ,
    "ethereal:micro_sakura_wood_15"                          ,
    "ethereal:micro_sakura_wood_2"                           ,
    "ethereal:micro_sakura_wood_4"                           ,
    "ethereal:micro_snowbrick"                               ,
    "ethereal:micro_snowbrick_1"                             ,
    "ethereal:micro_snowbrick_12"                            ,
    "ethereal:micro_snowbrick_14"                            ,
    "ethereal:micro_snowbrick_15"                            ,
    "ethereal:micro_snowbrick_2"                             ,
    "ethereal:micro_snowbrick_4"                             ,
    "ethereal:micro_willow_wood"                             ,
    "ethereal:micro_willow_wood_1"                           ,
    "ethereal:micro_willow_wood_12"                          ,
    "ethereal:micro_willow_wood_14"                          ,
    "ethereal:micro_willow_wood_15"                          ,
    "ethereal:micro_willow_wood_2"                           ,
    "ethereal:micro_willow_wood_4"                           ,
    "ethereal:micro_yellow_wood"                             ,
    "ethereal:micro_yellow_wood_1"                           ,
    "ethereal:micro_yellow_wood_12"                          ,
    "ethereal:micro_yellow_wood_14"                          ,
    "ethereal:micro_yellow_wood_15"                          ,
    "ethereal:micro_yellow_wood_2"                           ,
    "ethereal:micro_yellow_wood_4"                           ,

    "ethereal:mushroom"                                      ,
    "ethereal:mushroom_dirt"                                 ,
    "ethereal:mushroom_moss"                                 ,
    "ethereal:mushroom_pore"                                 ,
    "ethereal:mushroom_sapling"                              ,
    "ethereal:mushroom_trunk"                                ,

    "ethereal:olive"                                         ,
    "ethereal:olive_leaves"                                  ,
    "ethereal:olive_tree_sapling"                            ,
    "ethereal:olive_trunk"                                   ,
    "ethereal:olive_wood"                                    ,

    "ethereal:onion_1"                                       ,
    "ethereal:onion_2"                                       ,
    "ethereal:onion_3"                                       ,
    "ethereal:onion_4"                                       ,
    "ethereal:onion_5"                                       ,

    "ethereal:orange"                                        ,
    "ethereal:orange_leaves"                                 ,
    "ethereal:orange_tree_sapling"                           ,

    "ethereal:palm_sapling"                                  ,
    "ethereal:palm_trunk"                                    ,
    "ethereal:palm_wood"                                     ,
    "ethereal:palmleaves"                                    ,

    "ethereal:panel_bamboo_wood"                             ,
    "ethereal:panel_bamboo_wood_1"                           ,
    "ethereal:panel_bamboo_wood_12"                          ,
    "ethereal:panel_bamboo_wood_14"                          ,
    "ethereal:panel_bamboo_wood_15"                          ,
    "ethereal:panel_bamboo_wood_2"                           ,
    "ethereal:panel_bamboo_wood_4"                           ,
    "ethereal:panel_banana_wood"                             ,
    "ethereal:panel_banana_wood_1"                           ,
    "ethereal:panel_banana_wood_12"                          ,
    "ethereal:panel_banana_wood_14"                          ,
    "ethereal:panel_banana_wood_15"                          ,
    "ethereal:panel_banana_wood_2"                           ,
    "ethereal:panel_banana_wood_4"                           ,

    "ethereal:panel_birch_wood"                              ,
    "ethereal:panel_birch_wood_1"                            ,
    "ethereal:panel_birch_wood_12"                           ,
    "ethereal:panel_birch_wood_14"                           ,
    "ethereal:panel_birch_wood_15"                           ,
    "ethereal:panel_birch_wood_2"                            ,
    "ethereal:panel_birch_wood_4"                            ,

    "ethereal:panel_crystal_block"                           ,
    "ethereal:panel_crystal_block_1"                         ,
    "ethereal:panel_crystal_block_12"                        ,
    "ethereal:panel_crystal_block_14"                        ,
    "ethereal:panel_crystal_block_15"                        ,
    "ethereal:panel_crystal_block_2"                         ,
    "ethereal:panel_crystal_block_4"                         ,

    "ethereal:panel_dry_dirt"                                ,
    "ethereal:panel_dry_dirt_1"                              ,
    "ethereal:panel_dry_dirt_12"                             ,
    "ethereal:panel_dry_dirt_14"                             ,
    "ethereal:panel_dry_dirt_15"                             ,
    "ethereal:panel_dry_dirt_2"                              ,
    "ethereal:panel_dry_dirt_4"                              ,

    "ethereal:panel_frost_wood"                              ,
    "ethereal:panel_frost_wood_1"                            ,
    "ethereal:panel_frost_wood_12"                           ,
    "ethereal:panel_frost_wood_14"                           ,
    "ethereal:panel_frost_wood_15"                           ,
    "ethereal:panel_frost_wood_2"                            ,
    "ethereal:panel_frost_wood_4"                            ,

    "ethereal:panel_icebrick"                                ,
    "ethereal:panel_icebrick_1"                              ,
    "ethereal:panel_icebrick_12"                             ,
    "ethereal:panel_icebrick_14"                             ,
    "ethereal:panel_icebrick_15"                             ,
    "ethereal:panel_icebrick_2"                              ,
    "ethereal:panel_icebrick_4"                              ,

    "ethereal:panel_mushroom"                                ,
    "ethereal:panel_mushroom_1"                              ,
    "ethereal:panel_mushroom_12"                             ,
    "ethereal:panel_mushroom_14"                             ,
    "ethereal:panel_mushroom_15"                             ,
    "ethereal:panel_mushroom_2"                              ,
    "ethereal:panel_mushroom_4"                              ,
    "ethereal:panel_mushroom_trunk"                          ,
    "ethereal:panel_mushroom_trunk_1"                        ,
    "ethereal:panel_mushroom_trunk_12"                       ,
    "ethereal:panel_mushroom_trunk_14"                       ,
    "ethereal:panel_mushroom_trunk_15"                       ,
    "ethereal:panel_mushroom_trunk_2"                        ,
    "ethereal:panel_mushroom_trunk_4"                        ,

    "ethereal:panel_olive_wood"                              ,
    "ethereal:panel_olive_wood_1"                            ,
    "ethereal:panel_olive_wood_12"                           ,
    "ethereal:panel_olive_wood_14"                           ,
    "ethereal:panel_olive_wood_15"                           ,
    "ethereal:panel_olive_wood_2"                            ,
    "ethereal:panel_olive_wood_4"                            ,

    "ethereal:panel_palm_wood"                               ,
    "ethereal:panel_palm_wood_1"                             ,
    "ethereal:panel_palm_wood_12"                            ,
    "ethereal:panel_palm_wood_14"                            ,
    "ethereal:panel_palm_wood_15"                            ,
    "ethereal:panel_palm_wood_2"                             ,
    "ethereal:panel_palm_wood_4"                             ,

    "ethereal:panel_redwood_wood"                            ,
    "ethereal:panel_redwood_wood_1"                          ,
    "ethereal:panel_redwood_wood_12"                         ,
    "ethereal:panel_redwood_wood_14"                         ,
    "ethereal:panel_redwood_wood_15"                         ,
    "ethereal:panel_redwood_wood_2"                          ,
    "ethereal:panel_redwood_wood_4"                          ,

    "ethereal:panel_sakura_wood"                             ,
    "ethereal:panel_sakura_wood_1"                           ,
    "ethereal:panel_sakura_wood_12"                          ,
    "ethereal:panel_sakura_wood_14"                          ,
    "ethereal:panel_sakura_wood_15"                          ,
    "ethereal:panel_sakura_wood_2"                           ,
    "ethereal:panel_sakura_wood_4"                           ,

    "ethereal:panel_snowbrick"                               ,
    "ethereal:panel_snowbrick_1"                             ,
    "ethereal:panel_snowbrick_12"                            ,
    "ethereal:panel_snowbrick_14"                            ,
    "ethereal:panel_snowbrick_15"                            ,
    "ethereal:panel_snowbrick_2"                             ,
    "ethereal:panel_snowbrick_4"                             ,

    "ethereal:panel_willow_wood"                             ,
    "ethereal:panel_willow_wood_1"                           ,
    "ethereal:panel_willow_wood_12"                          ,
    "ethereal:panel_willow_wood_14"                          ,
    "ethereal:panel_willow_wood_15"                          ,
    "ethereal:panel_willow_wood_2"                           ,
    "ethereal:panel_willow_wood_4"                           ,

    "ethereal:panel_yellow_wood"                             ,
    "ethereal:panel_yellow_wood_1"                           ,
    "ethereal:panel_yellow_wood_12"                          ,
    "ethereal:panel_yellow_wood_14"                          ,
    "ethereal:panel_yellow_wood_15"                          ,
    "ethereal:panel_yellow_wood_2"                           ,
    "ethereal:panel_yellow_wood_4"                           ,

    "ethereal:prairie_dirt"                                  ,
    "ethereal:quicksand"                                     ,
    "ethereal:quicksand2"                                    ,

    "ethereal:redwood_leaves"                                ,
    "ethereal:redwood_sapling"                               ,
    "ethereal:redwood_trunk"                                 ,
    "ethereal:redwood_wood"                                  ,

    "ethereal:sakura_leaves"                                 ,
    "ethereal:sakura_leaves2"                                ,
    "ethereal:sakura_sapling"                                ,
    "ethereal:sakura_trunk"                                  ,
    "ethereal:sakura_wood"                                   ,

    "ethereal:sandy"                                         ,
    "ethereal:scorched_tree"                                 ,
    "ethereal:seaweed"                                       ,

    "ethereal:slab_bamboo_wood"                              ,
    "ethereal:slab_bamboo_wood_1"                            ,
    "ethereal:slab_bamboo_wood_14"                           ,
    "ethereal:slab_bamboo_wood_15"                           ,
    "ethereal:slab_bamboo_wood_2"                            ,
    "ethereal:slab_bamboo_wood_quarter"                      ,
    "ethereal:slab_bamboo_wood_three_quarter"                ,
    "ethereal:slab_bamboo_wood_three_sides"                  ,
    "ethereal:slab_bamboo_wood_three_sides_u"                ,
    "ethereal:slab_bamboo_wood_two_sides"                    ,

    "ethereal:slab_banana_wood"                              ,
    "ethereal:slab_banana_wood_1"                            ,
    "ethereal:slab_banana_wood_14"                           ,
    "ethereal:slab_banana_wood_15"                           ,
    "ethereal:slab_banana_wood_2"                            ,
    "ethereal:slab_banana_wood_quarter"                      ,
    "ethereal:slab_banana_wood_three_quarter"                ,
    "ethereal:slab_banana_wood_three_sides"                  ,
    "ethereal:slab_banana_wood_three_sides_u"                ,
    "ethereal:slab_banana_wood_two_sides"                    ,

    "ethereal:slab_birch_wood"                               ,
    "ethereal:slab_birch_wood_1"                             ,
    "ethereal:slab_birch_wood_14"                            ,
    "ethereal:slab_birch_wood_15"                            ,
    "ethereal:slab_birch_wood_2"                             ,
    "ethereal:slab_birch_wood_quarter"                       ,
    "ethereal:slab_birch_wood_three_quarter"                 ,
    "ethereal:slab_birch_wood_three_sides"                   ,
    "ethereal:slab_birch_wood_three_sides_u"                 ,
    "ethereal:slab_birch_wood_two_sides"                     ,

    "ethereal:slab_crystal_block"                            ,
    "ethereal:slab_crystal_block_1"                          ,
    "ethereal:slab_crystal_block_14"                         ,
    "ethereal:slab_crystal_block_15"                         ,
    "ethereal:slab_crystal_block_2"                          ,
    "ethereal:slab_crystal_block_quarter"                    ,
    "ethereal:slab_crystal_block_three_quarter"              ,
    "ethereal:slab_crystal_block_three_sides"                ,
    "ethereal:slab_crystal_block_three_sides_u"              ,
    "ethereal:slab_crystal_block_two_sides"                  ,

    "ethereal:slab_dry_dirt"                                 ,
    "ethereal:slab_dry_dirt_1"                               ,
    "ethereal:slab_dry_dirt_14"                              ,
    "ethereal:slab_dry_dirt_15"                              ,
    "ethereal:slab_dry_dirt_2"                               ,
    "ethereal:slab_dry_dirt_quarter"                         ,
    "ethereal:slab_dry_dirt_three_quarter"                   ,
    "ethereal:slab_dry_dirt_three_sides"                     ,
    "ethereal:slab_dry_dirt_three_sides_u"                   ,
    "ethereal:slab_dry_dirt_two_sides"                       ,

    "ethereal:slab_frost_wood"                               ,
    "ethereal:slab_frost_wood_1"                             ,
    "ethereal:slab_frost_wood_14"                            ,
    "ethereal:slab_frost_wood_15"                            ,
    "ethereal:slab_frost_wood_2"                             ,
    "ethereal:slab_frost_wood_quarter"                       ,
    "ethereal:slab_frost_wood_three_quarter"                 ,
    "ethereal:slab_frost_wood_three_sides"                   ,
    "ethereal:slab_frost_wood_three_sides_u"                 ,
    "ethereal:slab_frost_wood_two_sides"                     ,

    "ethereal:slab_icebrick"                                 ,
    "ethereal:slab_icebrick_1"                               ,
    "ethereal:slab_icebrick_14"                              ,
    "ethereal:slab_icebrick_15"                              ,
    "ethereal:slab_icebrick_2"                               ,
    "ethereal:slab_icebrick_quarter"                         ,
    "ethereal:slab_icebrick_three_quarter"                   ,
    "ethereal:slab_icebrick_three_sides"                     ,
    "ethereal:slab_icebrick_three_sides_u"                   ,
    "ethereal:slab_icebrick_two_sides"                       ,

    "ethereal:slab_mushroom"                                 ,
    "ethereal:slab_mushroom_1"                               ,
    "ethereal:slab_mushroom_14"                              ,
    "ethereal:slab_mushroom_15"                              ,
    "ethereal:slab_mushroom_2"                               ,
    "ethereal:slab_mushroom_quarter"                         ,
    "ethereal:slab_mushroom_three_quarter"                   ,
    "ethereal:slab_mushroom_three_sides"                     ,
    "ethereal:slab_mushroom_three_sides_u"                   ,
    "ethereal:slab_mushroom_trunk"                           ,
    "ethereal:slab_mushroom_trunk_1"                         ,
    "ethereal:slab_mushroom_trunk_14"                        ,
    "ethereal:slab_mushroom_trunk_15"                        ,
    "ethereal:slab_mushroom_trunk_2"                         ,
    "ethereal:slab_mushroom_trunk_quarter"                   ,
    "ethereal:slab_mushroom_trunk_three_quarter"             ,
    "ethereal:slab_mushroom_trunk_three_sides"               ,
    "ethereal:slab_mushroom_trunk_three_sides_u"             ,
    "ethereal:slab_mushroom_trunk_two_sides"                 ,
    "ethereal:slab_mushroom_two_sides"                       ,

    "ethereal:slab_olive_wood"                               ,
    "ethereal:slab_olive_wood_1"                             ,
    "ethereal:slab_olive_wood_14"                            ,
    "ethereal:slab_olive_wood_15"                            ,
    "ethereal:slab_olive_wood_2"                             ,
    "ethereal:slab_olive_wood_quarter"                       ,
    "ethereal:slab_olive_wood_three_quarter"                 ,
    "ethereal:slab_olive_wood_three_sides"                   ,
    "ethereal:slab_olive_wood_three_sides_u"                 ,
    "ethereal:slab_olive_wood_two_sides"                     ,

    "ethereal:slab_palm_wood"                                ,
    "ethereal:slab_palm_wood_1"                              ,
    "ethereal:slab_palm_wood_14"                             ,
    "ethereal:slab_palm_wood_15"                             ,
    "ethereal:slab_palm_wood_2"                              ,
    "ethereal:slab_palm_wood_quarter"                        ,
    "ethereal:slab_palm_wood_three_quarter"                  ,
    "ethereal:slab_palm_wood_three_sides"                    ,
    "ethereal:slab_palm_wood_three_sides_u"                  ,
    "ethereal:slab_palm_wood_two_sides"                      ,

    "ethereal:slab_redwood_wood"                             ,
    "ethereal:slab_redwood_wood_1"                           ,
    "ethereal:slab_redwood_wood_14"                          ,
    "ethereal:slab_redwood_wood_15"                          ,
    "ethereal:slab_redwood_wood_2"                           ,
    "ethereal:slab_redwood_wood_quarter"                     ,
    "ethereal:slab_redwood_wood_three_quarter"               ,
    "ethereal:slab_redwood_wood_three_sides"                 ,
    "ethereal:slab_redwood_wood_three_sides_u"               ,
    "ethereal:slab_redwood_wood_two_sides"                   ,

    "ethereal:slab_sakura_wood"                              ,
    "ethereal:slab_sakura_wood_1"                            ,
    "ethereal:slab_sakura_wood_14"                           ,
    "ethereal:slab_sakura_wood_15"                           ,
    "ethereal:slab_sakura_wood_2"                            ,
    "ethereal:slab_sakura_wood_quarter"                      ,
    "ethereal:slab_sakura_wood_three_quarter"                ,
    "ethereal:slab_sakura_wood_three_sides"                  ,
    "ethereal:slab_sakura_wood_three_sides_u"                ,
    "ethereal:slab_sakura_wood_two_sides"                    ,

    "ethereal:slab_snowbrick"                                ,
    "ethereal:slab_snowbrick_1"                              ,
    "ethereal:slab_snowbrick_14"                             ,
    "ethereal:slab_snowbrick_15"                             ,
    "ethereal:slab_snowbrick_2"                              ,
    "ethereal:slab_snowbrick_quarter"                        ,
    "ethereal:slab_snowbrick_three_quarter"                  ,
    "ethereal:slab_snowbrick_three_sides"                    ,
    "ethereal:slab_snowbrick_three_sides_u"                  ,
    "ethereal:slab_snowbrick_two_sides"                      ,

    "ethereal:slab_willow_wood"                              ,
    "ethereal:slab_willow_wood_1"                            ,
    "ethereal:slab_willow_wood_14"                           ,
    "ethereal:slab_willow_wood_15"                           ,
    "ethereal:slab_willow_wood_2"                            ,
    "ethereal:slab_willow_wood_quarter"                      ,
    "ethereal:slab_willow_wood_three_quarter"                ,
    "ethereal:slab_willow_wood_three_sides"                  ,
    "ethereal:slab_willow_wood_three_sides_u"                ,
    "ethereal:slab_willow_wood_two_sides"                    ,
    "ethereal:slab_yellow_wood"                              ,
    "ethereal:slab_yellow_wood_1"                            ,
    "ethereal:slab_yellow_wood_14"                           ,
    "ethereal:slab_yellow_wood_15"                           ,
    "ethereal:slab_yellow_wood_2"                            ,
    "ethereal:slab_yellow_wood_quarter"                      ,
    "ethereal:slab_yellow_wood_three_quarter"                ,
    "ethereal:slab_yellow_wood_three_sides"                  ,
    "ethereal:slab_yellow_wood_three_sides_u"                ,
    "ethereal:slab_yellow_wood_two_sides"                    ,

    "ethereal:slope_bamboo_wood"                             ,
    "ethereal:slope_bamboo_wood_cut"                         ,
    "ethereal:slope_bamboo_wood_half"                        ,
    "ethereal:slope_bamboo_wood_half_raised"                 ,
    "ethereal:slope_bamboo_wood_inner"                       ,
    "ethereal:slope_bamboo_wood_inner_cut"                   ,
    "ethereal:slope_bamboo_wood_inner_cut_half"              ,
    "ethereal:slope_bamboo_wood_inner_cut_half_raised"       ,
    "ethereal:slope_bamboo_wood_inner_half"                  ,
    "ethereal:slope_bamboo_wood_inner_half_raised"           ,
    "ethereal:slope_bamboo_wood_outer"                       ,
    "ethereal:slope_bamboo_wood_outer_cut"                   ,
    "ethereal:slope_bamboo_wood_outer_cut_half"              ,
    "ethereal:slope_bamboo_wood_outer_cut_half_raised"       ,
    "ethereal:slope_bamboo_wood_outer_half"                  ,
    "ethereal:slope_bamboo_wood_outer_half_raised"           ,

    "ethereal:slope_banana_wood"                             ,
    "ethereal:slope_banana_wood_cut"                         ,
    "ethereal:slope_banana_wood_half"                        ,
    "ethereal:slope_banana_wood_half_raised"                 ,
    "ethereal:slope_banana_wood_inner"                       ,
    "ethereal:slope_banana_wood_inner_cut"                   ,
    "ethereal:slope_banana_wood_inner_cut_half"              ,
    "ethereal:slope_banana_wood_inner_cut_half_raised"       ,
    "ethereal:slope_banana_wood_inner_half"                  ,
    "ethereal:slope_banana_wood_inner_half_raised"           ,
    "ethereal:slope_banana_wood_outer"                       ,
    "ethereal:slope_banana_wood_outer_cut"                   ,
    "ethereal:slope_banana_wood_outer_cut_half"              ,
    "ethereal:slope_banana_wood_outer_cut_half_raised"       ,
    "ethereal:slope_banana_wood_outer_half"                  ,
    "ethereal:slope_banana_wood_outer_half_raised"           ,

    "ethereal:slope_birch_wood"                              ,
    "ethereal:slope_birch_wood_cut"                          ,
    "ethereal:slope_birch_wood_half"                         ,
    "ethereal:slope_birch_wood_half_raised"                  ,
    "ethereal:slope_birch_wood_inner"                        ,
    "ethereal:slope_birch_wood_inner_cut"                    ,
    "ethereal:slope_birch_wood_inner_cut_half"               ,
    "ethereal:slope_birch_wood_inner_cut_half_raised"        ,
    "ethereal:slope_birch_wood_inner_half"                   ,
    "ethereal:slope_birch_wood_inner_half_raised"            ,
    "ethereal:slope_birch_wood_outer"                        ,
    "ethereal:slope_birch_wood_outer_cut"                    ,
    "ethereal:slope_birch_wood_outer_cut_half"               ,
    "ethereal:slope_birch_wood_outer_cut_half_raised"        ,
    "ethereal:slope_birch_wood_outer_half"                   ,
    "ethereal:slope_birch_wood_outer_half_raised"            ,

    "ethereal:slope_crystal_block"                           ,
    "ethereal:slope_crystal_block_cut"                       ,
    "ethereal:slope_crystal_block_half"                      ,
    "ethereal:slope_crystal_block_half_raised"               ,
    "ethereal:slope_crystal_block_inner"                     ,
    "ethereal:slope_crystal_block_inner_cut"                 ,
    "ethereal:slope_crystal_block_inner_cut_half"            ,
    "ethereal:slope_crystal_block_inner_cut_half_raised"     ,
    "ethereal:slope_crystal_block_inner_half"                ,
    "ethereal:slope_crystal_block_inner_half_raised"         ,
    "ethereal:slope_crystal_block_outer"                     ,
    "ethereal:slope_crystal_block_outer_cut"                 ,
    "ethereal:slope_crystal_block_outer_cut_half"            ,
    "ethereal:slope_crystal_block_outer_cut_half_raised"     ,
    "ethereal:slope_crystal_block_outer_half"                ,
    "ethereal:slope_crystal_block_outer_half_raised"         ,

    "ethereal:slope_dry_dirt"                                ,
    "ethereal:slope_dry_dirt_cut"                            ,
    "ethereal:slope_dry_dirt_half"                           ,
    "ethereal:slope_dry_dirt_half_raised"                    ,
    "ethereal:slope_dry_dirt_inner"                          ,
    "ethereal:slope_dry_dirt_inner_cut"                      ,
    "ethereal:slope_dry_dirt_inner_cut_half"                 ,
    "ethereal:slope_dry_dirt_inner_cut_half_raised"          ,
    "ethereal:slope_dry_dirt_inner_half"                     ,
    "ethereal:slope_dry_dirt_inner_half_raised"              ,
    "ethereal:slope_dry_dirt_outer"                          ,
    "ethereal:slope_dry_dirt_outer_cut"                      ,
    "ethereal:slope_dry_dirt_outer_cut_half"                 ,
    "ethereal:slope_dry_dirt_outer_cut_half_raised"          ,
    "ethereal:slope_dry_dirt_outer_half"                     ,
    "ethereal:slope_dry_dirt_outer_half_raised"              ,

    "ethereal:slope_frost_wood"                              ,
    "ethereal:slope_frost_wood_cut"                          ,
    "ethereal:slope_frost_wood_half"                         ,
    "ethereal:slope_frost_wood_half_raised"                  ,
    "ethereal:slope_frost_wood_inner"                        ,
    "ethereal:slope_frost_wood_inner_cut"                    ,
    "ethereal:slope_frost_wood_inner_cut_half"               ,
    "ethereal:slope_frost_wood_inner_cut_half_raised"        ,
    "ethereal:slope_frost_wood_inner_half"                   ,
    "ethereal:slope_frost_wood_inner_half_raised"            ,
    "ethereal:slope_frost_wood_outer"                        ,
    "ethereal:slope_frost_wood_outer_cut"                    ,
    "ethereal:slope_frost_wood_outer_cut_half"               ,
    "ethereal:slope_frost_wood_outer_cut_half_raised"        ,
    "ethereal:slope_frost_wood_outer_half"                   ,
    "ethereal:slope_frost_wood_outer_half_raised"            ,

    "ethereal:slope_icebrick"                                ,
    "ethereal:slope_icebrick_cut"                            ,
    "ethereal:slope_icebrick_half"                           ,
    "ethereal:slope_icebrick_half_raised"                    ,
    "ethereal:slope_icebrick_inner"                          ,
    "ethereal:slope_icebrick_inner_cut"                      ,
    "ethereal:slope_icebrick_inner_cut_half"                 ,
    "ethereal:slope_icebrick_inner_cut_half_raised"          ,
    "ethereal:slope_icebrick_inner_half"                     ,
    "ethereal:slope_icebrick_inner_half_raised"              ,
    "ethereal:slope_icebrick_outer"                          ,
    "ethereal:slope_icebrick_outer_cut"                      ,
    "ethereal:slope_icebrick_outer_cut_half"                 ,
    "ethereal:slope_icebrick_outer_cut_half_raised"          ,
    "ethereal:slope_icebrick_outer_half"                     ,
    "ethereal:slope_icebrick_outer_half_raised"              ,

    "ethereal:slope_mushroom"                                ,
    "ethereal:slope_mushroom_cut"                            ,
    "ethereal:slope_mushroom_half"                           ,
    "ethereal:slope_mushroom_half_raised"                    ,
    "ethereal:slope_mushroom_inner"                          ,
    "ethereal:slope_mushroom_inner_cut"                      ,
    "ethereal:slope_mushroom_inner_cut_half"                 ,
    "ethereal:slope_mushroom_inner_cut_half_raised"          ,
    "ethereal:slope_mushroom_inner_half"                     ,
    "ethereal:slope_mushroom_inner_half_raised"              ,
    "ethereal:slope_mushroom_outer"                          ,
    "ethereal:slope_mushroom_outer_cut"                      ,
    "ethereal:slope_mushroom_outer_cut_half"                 ,
    "ethereal:slope_mushroom_outer_cut_half_raised"          ,
    "ethereal:slope_mushroom_outer_half"                     ,
    "ethereal:slope_mushroom_outer_half_raised"              ,
    "ethereal:slope_mushroom_trunk"                          ,
    "ethereal:slope_mushroom_trunk_cut"                      ,
    "ethereal:slope_mushroom_trunk_half"                     ,
    "ethereal:slope_mushroom_trunk_half_raised"              ,
    "ethereal:slope_mushroom_trunk_inner"                    ,
    "ethereal:slope_mushroom_trunk_inner_cut"                ,
    "ethereal:slope_mushroom_trunk_inner_cut_half"           ,
    "ethereal:slope_mushroom_trunk_inner_cut_half_raised"    ,
    "ethereal:slope_mushroom_trunk_inner_half"               ,
    "ethereal:slope_mushroom_trunk_inner_half_raised"        ,
    "ethereal:slope_mushroom_trunk_outer"                    ,
    "ethereal:slope_mushroom_trunk_outer_cut"                ,
    "ethereal:slope_mushroom_trunk_outer_cut_half"           ,
    "ethereal:slope_mushroom_trunk_outer_cut_half_raised"    ,
    "ethereal:slope_mushroom_trunk_outer_half"               ,
    "ethereal:slope_mushroom_trunk_outer_half_raised"        ,

    "ethereal:slope_olive_wood"                              ,
    "ethereal:slope_olive_wood_cut"                          ,
    "ethereal:slope_olive_wood_half"                         ,
    "ethereal:slope_olive_wood_half_raised"                  ,
    "ethereal:slope_olive_wood_inner"                        ,
    "ethereal:slope_olive_wood_inner_cut"                    ,
    "ethereal:slope_olive_wood_inner_cut_half"               ,
    "ethereal:slope_olive_wood_inner_cut_half_raised"        ,
    "ethereal:slope_olive_wood_inner_half"                   ,
    "ethereal:slope_olive_wood_inner_half_raised"            ,
    "ethereal:slope_olive_wood_outer"                        ,
    "ethereal:slope_olive_wood_outer_cut"                    ,
    "ethereal:slope_olive_wood_outer_cut_half"               ,
    "ethereal:slope_olive_wood_outer_cut_half_raised"        ,
    "ethereal:slope_olive_wood_outer_half"                   ,
    "ethereal:slope_olive_wood_outer_half_raised"            ,

    "ethereal:slope_palm_wood"                               ,
    "ethereal:slope_palm_wood_cut"                           ,
    "ethereal:slope_palm_wood_half"                          ,
    "ethereal:slope_palm_wood_half_raised"                   ,
    "ethereal:slope_palm_wood_inner"                         ,
    "ethereal:slope_palm_wood_inner_cut"                     ,
    "ethereal:slope_palm_wood_inner_cut_half"                ,
    "ethereal:slope_palm_wood_inner_cut_half_raised"         ,
    "ethereal:slope_palm_wood_inner_half"                    ,
    "ethereal:slope_palm_wood_inner_half_raised"             ,
    "ethereal:slope_palm_wood_outer"                         ,
    "ethereal:slope_palm_wood_outer_cut"                     ,
    "ethereal:slope_palm_wood_outer_cut_half"                ,
    "ethereal:slope_palm_wood_outer_cut_half_raised"         ,
    "ethereal:slope_palm_wood_outer_half"                    ,
    "ethereal:slope_palm_wood_outer_half_raised"             ,

    "ethereal:slope_redwood_wood"                            ,
    "ethereal:slope_redwood_wood_cut"                        ,
    "ethereal:slope_redwood_wood_half"                       ,
    "ethereal:slope_redwood_wood_half_raised"                ,
    "ethereal:slope_redwood_wood_inner"                      ,
    "ethereal:slope_redwood_wood_inner_cut"                  ,
    "ethereal:slope_redwood_wood_inner_cut_half"             ,
    "ethereal:slope_redwood_wood_inner_cut_half_raised"      ,
    "ethereal:slope_redwood_wood_inner_half"                 ,
    "ethereal:slope_redwood_wood_inner_half_raised"          ,
    "ethereal:slope_redwood_wood_outer"                      ,
    "ethereal:slope_redwood_wood_outer_cut"                  ,
    "ethereal:slope_redwood_wood_outer_cut_half"             ,
    "ethereal:slope_redwood_wood_outer_cut_half_raised"      ,
    "ethereal:slope_redwood_wood_outer_half"                 ,
    "ethereal:slope_redwood_wood_outer_half_raised"          ,

    "ethereal:slope_sakura_wood"                             ,
    "ethereal:slope_sakura_wood_cut"                         ,
    "ethereal:slope_sakura_wood_half"                        ,
    "ethereal:slope_sakura_wood_half_raised"                 ,
    "ethereal:slope_sakura_wood_inner"                       ,
    "ethereal:slope_sakura_wood_inner_cut"                   ,
    "ethereal:slope_sakura_wood_inner_cut_half"              ,
    "ethereal:slope_sakura_wood_inner_cut_half_raised"       ,
    "ethereal:slope_sakura_wood_inner_half"                  ,
    "ethereal:slope_sakura_wood_inner_half_raised"           ,
    "ethereal:slope_sakura_wood_outer"                       ,
    "ethereal:slope_sakura_wood_outer_cut"                   ,
    "ethereal:slope_sakura_wood_outer_cut_half"              ,
    "ethereal:slope_sakura_wood_outer_cut_half_raised"       ,
    "ethereal:slope_sakura_wood_outer_half"                  ,
    "ethereal:slope_sakura_wood_outer_half_raised"           ,

    "ethereal:slope_snowbrick"                               ,
    "ethereal:slope_snowbrick_cut"                           ,
    "ethereal:slope_snowbrick_half"                          ,
    "ethereal:slope_snowbrick_half_raised"                   ,
    "ethereal:slope_snowbrick_inner"                         ,
    "ethereal:slope_snowbrick_inner_cut"                     ,
    "ethereal:slope_snowbrick_inner_cut_half"                ,
    "ethereal:slope_snowbrick_inner_cut_half_raised"         ,
    "ethereal:slope_snowbrick_inner_half"                    ,
    "ethereal:slope_snowbrick_inner_half_raised"             ,
    "ethereal:slope_snowbrick_outer"                         ,
    "ethereal:slope_snowbrick_outer_cut"                     ,
    "ethereal:slope_snowbrick_outer_cut_half"                ,
    "ethereal:slope_snowbrick_outer_cut_half_raised"         ,
    "ethereal:slope_snowbrick_outer_half"                    ,
    "ethereal:slope_snowbrick_outer_half_raised"             ,

    "ethereal:slope_willow_wood"                             ,
    "ethereal:slope_willow_wood_cut"                         ,
    "ethereal:slope_willow_wood_half"                        ,
    "ethereal:slope_willow_wood_half_raised"                 ,
    "ethereal:slope_willow_wood_inner"                       ,
    "ethereal:slope_willow_wood_inner_cut"                   ,
    "ethereal:slope_willow_wood_inner_cut_half"              ,
    "ethereal:slope_willow_wood_inner_cut_half_raised"       ,
    "ethereal:slope_willow_wood_inner_half"                  ,
    "ethereal:slope_willow_wood_inner_half_raised"           ,
    "ethereal:slope_willow_wood_outer"                       ,
    "ethereal:slope_willow_wood_outer_cut"                   ,
    "ethereal:slope_willow_wood_outer_cut_half"              ,
    "ethereal:slope_willow_wood_outer_cut_half_raised"       ,
    "ethereal:slope_willow_wood_outer_half"                  ,
    "ethereal:slope_willow_wood_outer_half_raised"           ,

    "ethereal:slope_yellow_wood"                             ,
    "ethereal:slope_yellow_wood_cut"                         ,
    "ethereal:slope_yellow_wood_half"                        ,
    "ethereal:slope_yellow_wood_half_raised"                 ,
    "ethereal:slope_yellow_wood_inner"                       ,
    "ethereal:slope_yellow_wood_inner_cut"                   ,
    "ethereal:slope_yellow_wood_inner_cut_half"              ,
    "ethereal:slope_yellow_wood_inner_cut_half_raised"       ,
    "ethereal:slope_yellow_wood_inner_half"                  ,
    "ethereal:slope_yellow_wood_inner_half_raised"           ,
    "ethereal:slope_yellow_wood_outer"                       ,
    "ethereal:slope_yellow_wood_outer_cut"                   ,
    "ethereal:slope_yellow_wood_outer_cut_half"              ,
    "ethereal:slope_yellow_wood_outer_cut_half_raised"       ,
    "ethereal:slope_yellow_wood_outer_half"                  ,
    "ethereal:slope_yellow_wood_outer_half_raised"           ,

    "ethereal:snowbrick"                                     ,
    "ethereal:snowygrass"                                    ,

    "ethereal:stair_bamboo_wood"                             ,
    "ethereal:stair_bamboo_wood_alt"                         ,
    "ethereal:stair_bamboo_wood_alt_1"                       ,
    "ethereal:stair_bamboo_wood_alt_2"                       ,
    "ethereal:stair_bamboo_wood_alt_4"                       ,
    "ethereal:stair_bamboo_wood_half"                        ,
    "ethereal:stair_bamboo_wood_inner"                       ,
    "ethereal:stair_bamboo_wood_outer"                       ,
    "ethereal:stair_bamboo_wood_right_half"                  ,

    "ethereal:stair_banana_wood"                             ,
    "ethereal:stair_banana_wood_alt"                         ,
    "ethereal:stair_banana_wood_alt_1"                       ,
    "ethereal:stair_banana_wood_alt_2"                       ,
    "ethereal:stair_banana_wood_alt_4"                       ,
    "ethereal:stair_banana_wood_half"                        ,
    "ethereal:stair_banana_wood_inner"                       ,
    "ethereal:stair_banana_wood_outer"                       ,
    "ethereal:stair_banana_wood_right_half"                  ,

    "ethereal:stair_birch_wood"                              ,
    "ethereal:stair_birch_wood_alt"                          ,
    "ethereal:stair_birch_wood_alt_1"                        ,
    "ethereal:stair_birch_wood_alt_2"                        ,
    "ethereal:stair_birch_wood_alt_4"                        ,
    "ethereal:stair_birch_wood_half"                         ,
    "ethereal:stair_birch_wood_inner"                        ,
    "ethereal:stair_birch_wood_outer"                        ,
    "ethereal:stair_birch_wood_right_half"                   ,

    "ethereal:stair_crystal_block"                           ,
    "ethereal:stair_crystal_block_alt"                       ,
    "ethereal:stair_crystal_block_alt_1"                     ,
    "ethereal:stair_crystal_block_alt_2"                     ,
    "ethereal:stair_crystal_block_alt_4"                     ,
    "ethereal:stair_crystal_block_half"                      ,
    "ethereal:stair_crystal_block_inner"                     ,
    "ethereal:stair_crystal_block_outer"                     ,
    "ethereal:stair_crystal_block_right_half"                ,

    "ethereal:stair_dry_dirt"                                ,
    "ethereal:stair_dry_dirt_alt"                            ,
    "ethereal:stair_dry_dirt_alt_1"                          ,
    "ethereal:stair_dry_dirt_alt_2"                          ,
    "ethereal:stair_dry_dirt_alt_4"                          ,
    "ethereal:stair_dry_dirt_half"                           ,
    "ethereal:stair_dry_dirt_inner"                          ,
    "ethereal:stair_dry_dirt_outer"                          ,
    "ethereal:stair_dry_dirt_right_half"                     ,

    "ethereal:stair_frost_wood"                              ,
    "ethereal:stair_frost_wood_alt"                          ,
    "ethereal:stair_frost_wood_alt_1"                        ,
    "ethereal:stair_frost_wood_alt_2"                        ,
    "ethereal:stair_frost_wood_alt_4"                        ,
    "ethereal:stair_frost_wood_half"                         ,
    "ethereal:stair_frost_wood_inner"                        ,
    "ethereal:stair_frost_wood_outer"                        ,
    "ethereal:stair_frost_wood_right_half"                   ,

    "ethereal:stair_icebrick"                                ,
    "ethereal:stair_icebrick_alt"                            ,
    "ethereal:stair_icebrick_alt_1"                          ,
    "ethereal:stair_icebrick_alt_2"                          ,
    "ethereal:stair_icebrick_alt_4"                          ,
    "ethereal:stair_icebrick_half"                           ,
    "ethereal:stair_icebrick_inner"                          ,
    "ethereal:stair_icebrick_outer"                          ,
    "ethereal:stair_icebrick_right_half"                     ,

    "ethereal:stair_mushroom"                                ,
    "ethereal:stair_mushroom_alt"                            ,
    "ethereal:stair_mushroom_alt_1"                          ,
    "ethereal:stair_mushroom_alt_2"                          ,
    "ethereal:stair_mushroom_alt_4"                          ,
    "ethereal:stair_mushroom_half"                           ,
    "ethereal:stair_mushroom_inner"                          ,
    "ethereal:stair_mushroom_outer"                          ,
    "ethereal:stair_mushroom_right_half"                     ,
    "ethereal:stair_mushroom_trunk"                          ,
    "ethereal:stair_mushroom_trunk_alt"                      ,
    "ethereal:stair_mushroom_trunk_alt_1"                    ,
    "ethereal:stair_mushroom_trunk_alt_2"                    ,
    "ethereal:stair_mushroom_trunk_alt_4"                    ,
    "ethereal:stair_mushroom_trunk_half"                     ,
    "ethereal:stair_mushroom_trunk_inner"                    ,
    "ethereal:stair_mushroom_trunk_outer"                    ,
    "ethereal:stair_mushroom_trunk_right_half"               ,

    "ethereal:stair_olive_wood"                              ,
    "ethereal:stair_olive_wood_alt"                          ,
    "ethereal:stair_olive_wood_alt_1"                        ,
    "ethereal:stair_olive_wood_alt_2"                        ,
    "ethereal:stair_olive_wood_alt_4"                        ,
    "ethereal:stair_olive_wood_half"                         ,
    "ethereal:stair_olive_wood_inner"                        ,
    "ethereal:stair_olive_wood_outer"                        ,
    "ethereal:stair_olive_wood_right_half"                   ,

    "ethereal:stair_palm_wood"                               ,
    "ethereal:stair_palm_wood_alt"                           ,
    "ethereal:stair_palm_wood_alt_1"                         ,
    "ethereal:stair_palm_wood_alt_2"                         ,
    "ethereal:stair_palm_wood_alt_4"                         ,
    "ethereal:stair_palm_wood_half"                          ,
    "ethereal:stair_palm_wood_inner"                         ,
    "ethereal:stair_palm_wood_outer"                         ,
    "ethereal:stair_palm_wood_right_half"                    ,

    "ethereal:stair_redwood_wood"                            ,
    "ethereal:stair_redwood_wood_alt"                        ,
    "ethereal:stair_redwood_wood_alt_1"                      ,
    "ethereal:stair_redwood_wood_alt_2"                      ,
    "ethereal:stair_redwood_wood_alt_4"                      ,
    "ethereal:stair_redwood_wood_half"                       ,
    "ethereal:stair_redwood_wood_inner"                      ,
    "ethereal:stair_redwood_wood_outer"                      ,
    "ethereal:stair_redwood_wood_right_half"                 ,

    "ethereal:stair_sakura_wood"                             ,
    "ethereal:stair_sakura_wood_alt"                         ,
    "ethereal:stair_sakura_wood_alt_1"                       ,
    "ethereal:stair_sakura_wood_alt_2"                       ,
    "ethereal:stair_sakura_wood_alt_4"                       ,
    "ethereal:stair_sakura_wood_half"                        ,
    "ethereal:stair_sakura_wood_inner"                       ,
    "ethereal:stair_sakura_wood_outer"                       ,
    "ethereal:stair_sakura_wood_right_half"                  ,

    "ethereal:stair_snowbrick"                               ,
    "ethereal:stair_snowbrick_alt"                           ,
    "ethereal:stair_snowbrick_alt_1"                         ,
    "ethereal:stair_snowbrick_alt_2"                         ,
    "ethereal:stair_snowbrick_alt_4"                         ,
    "ethereal:stair_snowbrick_half"                          ,
    "ethereal:stair_snowbrick_inner"                         ,
    "ethereal:stair_snowbrick_outer"                         ,
    "ethereal:stair_snowbrick_right_half"                    ,

    "ethereal:stair_willow_wood"                             ,
    "ethereal:stair_willow_wood_alt"                         ,
    "ethereal:stair_willow_wood_alt_1"                       ,
    "ethereal:stair_willow_wood_alt_2"                       ,
    "ethereal:stair_willow_wood_alt_4"                       ,
    "ethereal:stair_willow_wood_half"                        ,
    "ethereal:stair_willow_wood_inner"                       ,
    "ethereal:stair_willow_wood_outer"                       ,
    "ethereal:stair_willow_wood_right_half"                  ,

    "ethereal:stair_yellow_wood"                             ,
    "ethereal:stair_yellow_wood_alt"                         ,
    "ethereal:stair_yellow_wood_alt_1"                       ,
    "ethereal:stair_yellow_wood_alt_2"                       ,
    "ethereal:stair_yellow_wood_alt_4"                       ,
    "ethereal:stair_yellow_wood_half"                        ,
    "ethereal:stair_yellow_wood_inner"                       ,
    "ethereal:stair_yellow_wood_outer"                       ,
    "ethereal:stair_yellow_wood_right_half"                  ,

    "ethereal:stone_ladder"                                  ,

    "ethereal:strawberry_1"                                  ,
    "ethereal:strawberry_2"                                  ,
    "ethereal:strawberry_3"                                  ,
    "ethereal:strawberry_4"                                  ,
    "ethereal:strawberry_5"                                  ,
    "ethereal:strawberry_6"                                  ,
    "ethereal:strawberry_7"                                  ,
    "ethereal:strawberry_8"                                  ,

    "ethereal:vine"                                          ,

    "ethereal:willow_sapling"                                ,
    "ethereal:willow_trunk"                                  ,
    "ethereal:willow_twig"                                   ,
    "ethereal:willow_wood"                                   ,

    "ethereal:yellow_tree_sapling"                           ,
    "ethereal:yellow_trunk"                                  ,
    "ethereal:yellow_wood"                                   ,
    "ethereal:yellowleaves"                                  ,

    "fake_fire:embers"                                       ,
    "fake_fire:fancy_fire"                                   ,
    "fake_fire:ice_fire"                                     ,

    "farming:barley_1"                                       ,
    "farming:barley_2"                                       ,
    "farming:barley_3"                                       ,
    "farming:barley_4"                                       ,
    "farming:barley_5"                                       ,
    "farming:barley_6"                                       ,
    "farming:barley_7"                                       ,
    "farming:beanbush"                                       ,

    "farming:beanpole"                                       ,
    "farming:beanpole_1"                                     ,
    "farming:beanpole_2"                                     ,
    "farming:beanpole_3"                                     ,
    "farming:beanpole_4"                                     ,
    "farming:beanpole_5"                                     ,

    "farming:blueberry_1"                                    ,
    "farming:blueberry_2"                                    ,
    "farming:blueberry_3"                                    ,
    "farming:blueberry_4"                                    ,
    "farming:bottle_ethanol"                                 ,

    "farming:carrot_1"                                       ,
    "farming:carrot_2"                                       ,
    "farming:carrot_3"                                       ,
    "farming:carrot_4"                                       ,
    "farming:carrot_5"                                       ,
    "farming:carrot_6"                                       ,
    "farming:carrot_7"                                       ,
    "farming:carrot_8"                                       ,

    "farming:chili_1"                                        ,
    "farming:chili_2"                                        ,
    "farming:chili_3"                                        ,
    "farming:chili_4"                                        ,
    "farming:chili_5"                                        ,
    "farming:chili_6"                                        ,
    "farming:chili_7"                                        ,
    "farming:chili_8"                                        ,

    "farming:cocoa_1"                                        ,
    "farming:cocoa_2"                                        ,
    "farming:cocoa_3"                                        ,

    "farming:coffee_1"                                       ,
    "farming:coffee_2"                                       ,
    "farming:coffee_3"                                       ,
    "farming:coffee_4"                                       ,
    "farming:coffee_5"                                       ,
    "farming:coffee_cup"                                     ,
    "farming:coffee_cup_hot"                                 ,

    "farming:corn_1"                                         ,
    "farming:corn_2"                                         ,
    "farming:corn_3"                                         ,
    "farming:corn_4"                                         ,
    "farming:corn_5"                                         ,
    "farming:corn_6"                                         ,
    "farming:corn_7"                                         ,
    "farming:corn_8"                                         ,

    "farming:cotton_1"                                       ,
    "farming:cotton_2"                                       ,
    "farming:cotton_3"                                       ,
    "farming:cotton_4"                                       ,
    "farming:cotton_5"                                       ,
    "farming:cotton_6"                                       ,
    "farming:cotton_7"                                       ,
    "farming:cotton_8"                                       ,

    "farming:cucumber_1"                                     ,
    "farming:cucumber_2"                                     ,
    "farming:cucumber_3"                                     ,
    "farming:cucumber_4"                                     ,

    "farming:desert_sand_soil"                               ,
    "farming:desert_sand_soil_wet"                           ,
    "farming:drinking_cup"                                   ,
    "farming:grapebush"                                      ,

    "farming:grapes_1"                                       ,
    "farming:grapes_2"                                       ,
    "farming:grapes_3"                                       ,
    "farming:grapes_4"                                       ,
    "farming:grapes_5"                                       ,
    "farming:grapes_6"                                       ,
    "farming:grapes_7"                                       ,
    "farming:grapes_8"                                       ,

    "farming:hemp_1"                                         ,
    "farming:hemp_2"                                         ,
    "farming:hemp_3"                                         ,
    "farming:hemp_4"                                         ,
    "farming:hemp_5"                                         ,
    "farming:hemp_6"                                         ,
    "farming:hemp_7"                                         ,
    "farming:hemp_8"                                         ,
    "farming:hemp_oil"                                       ,
    "farming:hemp_rope"                                      ,

    "farming:jackolantern"                                   ,
    "farming:jackolantern_on"                                ,

    "farming:melon_1"                                        ,
    "farming:melon_2"                                        ,
    "farming:melon_3"                                        ,
    "farming:melon_4"                                        ,
    "farming:melon_5"                                        ,
    "farming:melon_6"                                        ,
    "farming:melon_7"                                        ,
    "farming:melon_8"                                        ,

    "farming:potato_1"                                       ,
    "farming:potato_2"                                       ,
    "farming:potato_3"                                       ,
    "farming:potato_4"                                       ,

    "farming:pumpkin"                                        ,
    "farming:pumpkin_1"                                      ,
    "farming:pumpkin_2"                                      ,
    "farming:pumpkin_3"                                      ,
    "farming:pumpkin_4"                                      ,
    "farming:pumpkin_5"                                      ,
    "farming:pumpkin_6"                                      ,
    "farming:pumpkin_7"                                      ,
    "farming:pumpkin_8"                                      ,

    "farming:raspberry_1"                                    ,
    "farming:raspberry_2"                                    ,
    "farming:raspberry_3"                                    ,
    "farming:raspberry_4"                                    ,

    "farming:rhubarb_1"                                      ,
    "farming:rhubarb_2"                                      ,
    "farming:rhubarb_3"                                      ,

    "farming:seed_barley"                                    ,
    "farming:seed_cotton"                                    ,
    "farming:seed_hemp"                                      ,
    "farming:seed_wheat"                                     ,
    "farming:soil"                                           ,
    "farming:soil_wet"                                       ,
    "farming:straw"                                          ,

    "farming:tomato_1"                                       ,
    "farming:tomato_2"                                       ,
    "farming:tomato_3"                                       ,
    "farming:tomato_4"                                       ,
    "farming:tomato_5"                                       ,
    "farming:tomato_6"                                       ,
    "farming:tomato_7"                                       ,
    "farming:tomato_8"                                       ,
    "farming:trellis"                                        ,

    "farming:wheat_1"                                        ,
    "farming:wheat_2"                                        ,
    "farming:wheat_3"                                        ,
    "farming:wheat_4"                                        ,
    "farming:wheat_5"                                        ,
    "farming:wheat_6"                                        ,
    "farming:wheat_7"                                        ,
    "farming:wheat_8"                                        ,

    "fire:basic_flame"                                       ,
    "fire:permanent_flame"                                   ,

    "flowers:dandelion_yellow"                               ,
    "flowers:mushroom_brown"                                 ,
    "flowers:mushroom_red"                                   ,
    "flowers:rose"                                           ,
    "flowers:viola"                                          ,
    "flowers:waterlily"                                      ,

    "handholds:climbable_air"                                ,
    "handholds:desert_stone"                                 ,
    "handholds:ice"                                          ,
    "handholds:sandstone"                                    ,
    "handholds:stone"                                        ,

    "homedecor:armchair"                                     ,
    "homedecor:bathroom_tiles_dark"                          ,
    "homedecor:bathroom_tiles_light"                         ,
    "homedecor:bathroom_tiles_medium"                        ,
    "homedecor:bed_kingsize"                                 ,
    "homedecor:bed_regular"                                  ,
    "homedecor:beer_mug"                                     ,
    "homedecor:beer_tap"                                     ,
    "homedecor:bench_large_2"                                ,
    "homedecor:bottle_green"                                 ,
    "homedecor:ceiling_lantern"                              ,
    "homedecor:coat_tree"                                    ,

    "homedecor:cobweb_centered"                              ,
    "homedecor:cobweb_corner"                                ,
    "homedecor:cobweb_flat"                                  ,
    "homedecor:cobweb_plantlike"                             ,
    "homedecor:coffee_maker"                                 ,
    "homedecor:curtain_closed"                               ,
    "homedecor:curtain_open"                                 ,
    "homedecor:curtainrod_wood"                              ,
    "homedecor:cutlery_set"                                  ,

    "homedecor:desk_globe"                                   ,
    "homedecor:dishwasher_steel"                             ,
    "homedecor:fence_wrought_iron"                           ,
    "homedecor:fence_wrought_iron_2"                         ,
    "homedecor:glowlight_half"                               ,
    "homedecor:glowlight_quarter"                            ,
    "homedecor:glowlight_small_cube"                         ,
    "homedecor:grandfather_clock"                            ,
    "homedecor:ground_lantern"                               ,
    "homedecor:hanging_lantern"                              ,

    "homedecor:japanese_wall_bottom"                         ,
    "homedecor:japanese_wall_middle"                         ,
    "homedecor:japanese_wall_top"                            ,
    "homedecor:kitchen_chair_wood"                           ,
    "homedecor:kitchen_faucet"                               ,
    "homedecor:lattice_lantern_large"                        ,
    "homedecor:lattice_lantern_small"                        ,
    "homedecor:office_chair_basic"                           ,
    "homedecor:piano"                                        ,
    "homedecor:plasma_lamp"                                  ,
    "homedecor:pool_table"                                   ,
    "homedecor:potted_geranium"                              ,
    "homedecor:potted_tulip"                                 ,

    "homedecor:shingle_inner_corner_asphalt"                 ,
    "homedecor:shingle_outer_corner_asphalt"                 ,
    "homedecor:shingle_side_asphalt"                         ,
    "homedecor:shingle_side_wood"                            ,
    "homedecor:shingles_wood"                                ,
    "homedecor:shower_head"                                  ,

    "homedecor:shrubbery_green"                              ,
    "homedecor:shrubbery_large_green"                        ,
    "homedecor:shrubbery_large_red"                          ,
    "homedecor:shrubbery_large_yellow"                       ,
    "homedecor:shrubbery_red"                                ,
    "homedecor:shrubbery_yellow"                             ,

    "homedecor:sink"                                         ,
    "homedecor:skylight"                                     ,
    "homedecor:skylight_frosted"                             ,
    "homedecor:stained_glass"                                ,
    "homedecor:stonepath"                                    ,
    "homedecor:table_lamp_hi"                                ,
    "homedecor:table_mahogany"                               ,
    "homedecor:taps"                                         ,
    "homedecor:tatami_mat"                                   ,

    "homedecor:toilet"                                       ,
    "homedecor:toilet_paper"                                 ,
    "homedecor:torch_wall"                                   ,
    "homedecor:trash_can"                                    ,
    "homedecor:trash_can_green"                              ,
    "homedecor:wall_sconce"                                  ,
    "homedecor:welcome_mat_green"                            ,
    "homedecor:window_plain"                                 ,

    "ilights:light"                                          ,
    "King998_HD1"                                            ,

    "lapis:lapis_block"                                      ,
    "lapis:lapis_brick"                                      ,
    "lapis:lapis_cobble"                                     ,
    "lapis:lapis_tile"                                       ,

    "lapis:lazurite_block"                                   ,
    "lapis:lazurite_brick"                                   ,
    "lapis:pyrite_block"                                     ,
    "lapis:pyrite_ore"                                       ,
    "lapis:rosace"                                           ,

    "lavalamp:lavalamp"                                      ,
    "lavalamp:lavalamp_off"                                  ,

    "lrfurn:armchair"                                        ,
    "lrfurn:coffeetable"                                     ,
    "lrfurn:endtable"                                        ,
    "lrfurn:longsofa"                                        ,
    "lrfurn:sofa"                                            ,

    "maptools:black"                                         ,
    "maptools:brick"                                         ,
    "maptools:bronzeblock"                                   ,
    "maptools:cactus"                                        ,
    "maptools:clay"                                          ,
    "maptools:climb"                                         ,
    "maptools:coalblock"                                     ,
    "maptools:cobble"                                        ,
    "maptools:copperblock"                                   ,

    "maptools:desert_cobble"                                 ,
    "maptools:desert_sand"                                   ,
    "maptools:desert_sand_soil_wet"                          ,
    "maptools:desert_stone"                                  ,
    "maptools:desert_stonebrick"                             ,
    "maptools:diamondblock"                                  ,
    "maptools:dirt"                                          ,

    "maptools:fake_fire"                                     ,
    "maptools:fake_walkable"                                 ,
    "maptools:fake_walkable_pointable"                       ,
    "maptools:fullclip"                                      ,
    "maptools:fullclip_face"                                 ,
    "maptools:fullgrass"                                     ,
    "maptools:glass"                                         ,
    "maptools:goldblock"                                     ,
    "maptools:grass"                                         ,
    "maptools:gravel"                                        ,
    "maptools:igniter"                                       ,

    "maptools:ignore_like"                                   ,
    "maptools:ignore_like_no_clip"                           ,
    "maptools:ignore_like_no_clip_no_point"                  ,
    "maptools:ignore_like_no_point"                          ,

    "maptools:jungletree"                                    ,
    "maptools:junglewood"                                    ,
    "maptools:kill"                                          ,
    "maptools:ladder"                                        ,
    "maptools:leaves"                                        ,
    "maptools:lightbulb"                                     ,
    "maptools:mossycobble"                                   ,
    "maptools:nobuild"                                       ,
    "maptools:nointeract"                                    ,
    "maptools:papyrus"                                       ,
    "maptools:permanent_fire"                                ,
    "maptools:playerclip"                                    ,
    "maptools:playerclip_bottom"                             ,
    "maptools:playerclip_top"                                ,
    "maptools:sand"                                          ,
    "maptools:sandstone"                                     ,
    "maptools:sandstone_brick"                               ,
    "maptools:smoke"                                         ,
    "maptools:soil_wet"                                      ,
    "maptools:steelblock"                                    ,
    "maptools:stone"                                         ,
    "maptools:stonebrick"                                    ,
    "maptools:superapple"                                    ,
    "maptools:tree"                                          ,
    "maptools:white"                                         ,
    "maptools:wood"                                          ,

    "mese_crystals:mese_crystal_ore1"                        ,
    "mese_crystals:mese_crystal_ore2"                        ,
    "mese_crystals:mese_crystal_ore3"                        ,
    "mese_crystals:mese_crystal_ore4"                        ,

    "mobs:cheeseblock"                                       ,
    "mobs:cobweb"                                            ,
    "mobs:egg"                                               ,
    "mobs:fence_wood"                                        ,
    "mobs:honey_block"                                       ,

    "moreblocks:glow_glass"                                  ,
    "moreblocks:super_glow_glass"                            ,

    "more_gold:legendary_gold"                               ,
    "more_gold:purified_gold"                                ,
    "more_gold:radiant_gold"                                 ,
    "more_gold:refined_gold"                                 ,
    "more_gold:true_gold"                                    ,

    "moreores:copper_rail"                                   ,
    "moreores:mineral_mithril"                               ,
    "moreores:mineral_silver"                                ,

    "mymillwork:base_fancy_default_sandstone"                ,
    "mymillwork:column_base_default_sandstone"               ,
    "mymillwork:column_default_sandstone"                    ,
    "mymillwork:column_quarter_default_sandstone"            ,
    "mymillwork:column_quarter_fancybase_default_sandstone"  ,
    "mymillwork:crownmould_default_sandstone"                ,
    "mymillwork:crownmould_ic_default_sandstone"             ,

    "nftools:aquamarine_ore"                                 ,
    "nyancat:nyancat"                                        ,
    "nyancat:nyancat_rainbow"                                ,

    "octutool:ladder"                                        ,
    "octutool:mese"                                          ,
    "octutool:string"                                        ,
    "octutool:string_shards"                                 ,

    "pbj_pup:pbj_pup"                                        ,
    "pie:choc_3"                                             ,
    "pie:pie_0"                                              ,
    "plasmascreen:stand"                                     ,
    "plasmascreen:tv"                                        ,
    "plasmascreen:tv_off"                                    ,
    "quartz:quartz_ore"                                      ,

    "riesenpilz:brown"                                       ,
    "riesenpilz:brown45"                                     ,
    "riesenpilz:fly_agaric"                                  ,
    "riesenpilz:glowshroom"                                  ,
    "riesenpilz:ground"                                      ,

    "riesenpilz:head_binge"                                  ,
    "riesenpilz:head_blue"                                   ,
    "riesenpilz:head_blue_bright"                            ,
    "riesenpilz:head_brown"                                  ,
    "riesenpilz:head_brown_bright"                           ,
    "riesenpilz:head_brown_full"                             ,
    "riesenpilz:head_orange"                                 ,
    "riesenpilz:head_red"                                    ,
    "riesenpilz:head_red_side"                               ,
    "riesenpilz:head_white"                                  ,
    "riesenpilz:head_yellow"                                 ,

    "riesenpilz:lamellas"                                    ,
    "riesenpilz:lavashroom"                                  ,
    "riesenpilz:nether_shroom"                               ,
    "riesenpilz:parasol"                                     ,
    "riesenpilz:red"                                         ,
    "riesenpilz:red45"                                       ,
    "riesenpilz:stem"                                        ,
    "riesenpilz:stem_blue"                                   ,
    "riesenpilz:stem_brown"                                  ,

    "stairs:slab_bakedclay_white"                            ,
    "stairs:slab_cobble"                                     ,
    "stairs:slab_glass"                                      ,
    "stairs:slab_hard_clay"                                  ,
    "stairs:slab_junglewood"                                 ,
    "stairs:slab_moonbrick"                                  ,
    "stairs:slab_mossycobble"                                ,
    "stairs:slab_pine_wood"                                  ,
    "stairs:slab_red"                                        ,
    "stairs:slab_silver_sandstone_block"                     ,
    "stairs:slab_stone"                                      ,
    "stairs:slab_stone_block"                                ,
    "stairs:slab_stonebrick"                                 ,
    "stairs:slab_stone_tile"                                 ,
    "stairs:slab_white"                                      ,
    "stairs:slab_wood"                                       ,

    "stairs:stair_bakedclay_red"                             ,
    "stairs:stair_bakedclay_white"                           ,
    "stairs:stair_brick"                                     ,
    "stairs:stair_cobble"                                    ,
    "stairs:stair_hard_clay"                                 ,
    "stairs:stair_inner_cobble"                              ,
    "stairs:stair_inner_junglewood"                          ,
    "stairs:stair_inner_stonebrick"                          ,
    "stairs:stair_inner_wood"                                ,
    "stairs:stair_junglewood"                                ,
    "stairs:stair_moonbrick"                                 ,
    "stairs:stair_mossycobble"                               ,
    "stairs:stair_outer_moonbrick"                           ,
    "stairs:stair_outer_stonebrick"                          ,
    "stairs:stair_outer_wood"                                ,
    "stairs:stair_pine_wood"                                 ,
    "stairs:stair_stonebrick"                                ,
    "stairs:stair_stone_tile"                                ,
    "stairs:stair_tree"                                      ,
    "stairs:stair_white"                                     ,
    "stairs:stair_wood"                                      ,

    "tnt:boom"                                               ,
    "tnt:gunpowder"                                          ,
    "tnt:gunpowder_burning"                                  ,

    "vessels:drinking_glass"                                 ,
    "vessels:glass_bottle"                                   ,
    "vessels:steel_bottle"                                   ,

    "walls_all:stone_brick"                                  ,
    "walls:cobble"                                           ,
    "walls:mossycobble"                                      ,

    "wool:blue"                                              ,
    "wool:dark_green"                                        ,
    "wool:green"                                             ,
    "wool:orange"                                            ,
    "wool:red"                                               ,
    "wool:white"                                             ,
    "wool:yellow"                                            ,

    "wylewny1"                                               ,

    "xdecor:candle"                                          ,
    "xdecor:cauldron_idle"                                   ,
    "xdecor:chair"                                           ,
    "xdecor:coalstone_tile"                                  ,
    "xdecor:cobweb"                                          ,
    "xdecor:hard_clay"                                       ,
    "xdecor:iron_lightbox"                                   ,
    "xdecor:ivy"                                             ,
    "xdecor:moonbrick"                                       ,

    "xdecor:potted_dandelion_white"                          ,
    "xdecor:potted_geranium"                                 ,
    "xdecor:potted_rose"                                     ,
    "xdecor:potted_tulip"                                    ,
    "xdecor:potted_viola"                                    ,

    "xdecor:stone_rune"                                      ,
    "xdecor:stone_tile"                                      ,
    "xdecor:table"                                           ,
    "xdecor:wooden_lightbox"                                 ,
    "xdecor:woodframed_glass"                                ,
    "xdecor:workbench"                                       ,
    "xpanes:bar_flat"                                        ,
    "xpanes:chainlink"                                       ,
    "xpanes:chainlink_flat"                                  ,
    "xpanes:pane_flat"                                       ,
}

-- ===================================================================

function worldedit.serialize(pos1, pos2)
    pos1, pos2 = worldedit.sort_pos(pos1, pos2)
    worldedit.keep_loaded(pos1, pos2)

    local pos = {x=pos1.x, y=0, z=0}
    local count = 0
    local result = {}
    local get_node, get_meta = minetest.get_node, minetest.get_meta

    while pos.x <= pos2.x do
        pos.y = pos1.y
        while pos.y <= pos2.y do
            pos.z = pos1.z
            while pos.z <= pos2.z do
                local node = get_node(pos)
                local nname  = node.name
                local copyit = 0

                if small_bowl == 0 then
                    if nname ~= "air" and nname ~= "ignore" then
                        copyit = 1
                    end
                else
                    if rejectwe [nname] == nil then copyit = 1 end
                end

                if copyit == 1 then
                    count = count + 1
                    local meta = get_meta(pos):to_table()

                    local meta_empty = true
                    -- Convert metadata item stacks to item strings
                    for name, inventory in pairs(meta.inventory) do
                        for index, stack in ipairs(inventory) do
                            meta_empty = false
                            inventory[index] = stack.to_string and stack:to_string() or stack
                        end
                    end
                    for k in pairs(meta) do
                        if k ~= "inventory" then
                            meta_empty = false
                            break
                        end
                    end

                    result[count] = {
                        x = pos.x - pos1.x,
                        y = pos.y - pos1.y,
                        z = pos.z - pos1.z,
                        name = node.name,
                        param1 = node.param1 ~= 0 and node.param1 or nil,
                        param2 = node.param2 ~= 0 and node.param2 or nil,
                        meta = not meta_empty and meta or nil,
                    }
                end
                pos.z = pos.z + 1
            end
            pos.y = pos.y + 1
        end
        pos.x = pos.x + 1
    end
    -- Serialize entries
    result = minetest.serialize(result)
    return LATEST_SERIALIZATION_HEADER .. result, count
end

-- ===================================================================

function worldedit.cerealize (pos1, pos2)
    small_bowl = 1
    local  result, count = worldedit.serialize (pos1, pos2)
    small_bowl = 0
    return result, count
end

-- ===================================================================

--- Loads the schematic in `value` into a node list in the latest format.
-- Contains code based on [table.save/table.load](http://lua-users.org/wiki/SaveTableToFile)
-- by ChillCode, available under the MIT license.
-- @return A node list in the latest format, or nil on failure.
local function load_schematic(value)
    local version, header, content = worldedit.read_header(value)
    local nodes = {}
    if version == 1 or version == 2 then -- Original flat table format
        local tables = minetest.deserialize(content)
        if not tables then return nil end

        -- Transform the node table into an array of nodes
        for i = 1, #tables do
            for j, v in pairs(tables[i]) do
                if type(v) == "table" then
                    tables[i][j] = tables[v[1]]
                end
            end
        end
        nodes = tables[1]

        if version == 1 then --original flat table format
            for i, entry in ipairs(nodes) do
                local pos = entry[1]
                entry.x, entry.y, entry.z = pos.x, pos.y, pos.z
                entry[1] = nil
                local node = entry[2]
                entry.name, entry.param1, entry.param2 = node.name, node.param1, node.param2
                entry[2] = nil
            end
        end
    elseif version == 3 then -- List format
        for x, y, z, name, param1, param2 in content:gmatch(
                "([+-]?%d+)%s+([+-]?%d+)%s+([+-]?%d+)%s+" ..
                "([^%s]+)%s+(%d+)%s+(%d+)[^\r\n]*[\r\n]*") do
            param1, param2 = tonumber(param1), tonumber(param2)
            table.insert(nodes, {
                x = tonumber(x),
                y = tonumber(y),
                z = tonumber(z),
                name = name,
                param1 = param1 ~= 0 and param1 or nil,
                param2 = param2 ~= 0 and param2 or nil,
            })
        end
    elseif version == 4 or version == 5 then -- Nested table format
        if not jit then
            -- This is broken for larger tables in the current version of LuaJIT
            nodes = minetest.deserialize(content)
        else
            -- XXX: This is a filthy hack that works surprisingly well - in LuaJIT, `minetest.deserialize` will fail due to the register limit
            nodes = {}
            content = content:gsub("return%s*{", "", 1):gsub("}%s*$", "", 1) -- remove the starting and ending values to leave only the node data
            local escaped = content:gsub("\\\\", "@@"):gsub("\\\"", "@@"):gsub("(\"[^\"]*\")", function(s) return string.rep("@", #s) end)
            local startpos, startpos1, endpos = 1, 1
            while true do -- go through each individual node entry (except the last)
                startpos, endpos = escaped:find("},%s*{", startpos)
                if not startpos then
                    break
                end
                local current = content:sub(startpos1, startpos)
                local entry = minetest.deserialize("return " .. current)
                table.insert(nodes, entry)
                startpos, startpos1 = endpos, endpos
            end
            local entry = minetest.deserialize("return " .. content:sub(startpos1)) -- process the last entry
            table.insert(nodes, entry)
        end
    else
        return nil
    end
    return nodes
end

--- Determines the volume the nodes represented by string `value` would occupy
-- if deserialized at `origin_pos`.
-- @return Low corner position.
-- @return High corner position.
-- @return The number of nodes.
function worldedit.allocate(origin_pos, value)
    local nodes = load_schematic(value)
    if not nodes then return nil end
    return worldedit.allocate_with_nodes(origin_pos, nodes)
end


-- Internal
function worldedit.allocate_with_nodes(origin_pos, nodes)
    local huge = math.huge
    local pos1x, pos1y, pos1z = huge, huge, huge
    local pos2x, pos2y, pos2z = -huge, -huge, -huge
    local origin_x, origin_y, origin_z = origin_pos.x, origin_pos.y, origin_pos.z
    for i, entry in ipairs(nodes) do
        local x, y, z = origin_x + entry.x, origin_y + entry.y, origin_z + entry.z
        if x < pos1x then pos1x = x end
        if y < pos1y then pos1y = y end
        if z < pos1z then pos1z = z end
        if x > pos2x then pos2x = x end
        if y > pos2y then pos2y = y end
        if z > pos2z then pos2z = z end
    end
    local pos1 = {x=pos1x, y=pos1y, z=pos1z}
    local pos2 = {x=pos2x, y=pos2y, z=pos2z}
    return pos1, pos2, #nodes
end


--- Loads the nodes represented by string `value` at position `origin_pos`.
-- @return The number of nodes deserialized.
function worldedit.deserialize(origin_pos, value)
    local nodes = load_schematic(value)
    if not nodes then return nil end

    local pos1, pos2 = worldedit.allocate_with_nodes(origin_pos, nodes)
    worldedit.keep_loaded(pos1, pos2)

    local origin_x, origin_y, origin_z = origin_pos.x, origin_pos.y, origin_pos.z
    local count = 0
    local add_node, get_meta = minetest.add_node, minetest.get_meta
    for i, entry in ipairs(nodes) do
        entry.x, entry.y, entry.z = origin_x + entry.x, origin_y + entry.y, origin_z + entry.z
        -- Entry acts as both position and node
        add_node(entry, entry)
        if entry.meta then
            get_meta(entry):from_table(entry.meta)
        end
    end
    return #nodes
end
